import processing.opengl.*;
import processing.pdf.*;
import java.util.Calendar;
    
int lineCount, _lineCount = 7; // per 100px
int lineDefinition = 8;
float lineWeight = 1;
float maxLineWeightCoeff = 4;
float maxWaveHeight = 500; // z-height
float maxWaveRadius, _maxWaveRadius = 0.7; // percentage of window width
int waveCount = 300;
float lineHeightImperfectionRange = 0.2; // of lineSpacing
float waveVelocity = 0.01;
float rotationX = 0.7, rotationY = 0, rotationZ = 0.05;

float lineSpacing;
float[][] lineHeightImperfections;
float[] flatteningDistances;
float flatteningDistancesSum = 0;
boolean savePicture = false;
float maxLineWeight = lineWeight * maxLineWeightCoeff;
Wave[] waves = new Wave[waveCount];

void setup() {
  size(1920, 1080, P3D);
  smooth(9);
  background(255);
   
  lineCount = floor(_lineCount * width / 100);
  lineSpacing = height / lineCount;
  lineDefinition = int(lineDefinition * width / 100);
  if (width < height) lineHeightImperfectionRange = 0.5;
  maxWaveRadius = _maxWaveRadius * width;

  lineHeightImperfections = computeLineHeightImperfections();
  flatteningDistances = computeFlatteningDistances();
  
  for (int i = 0; i < waveCount; i++)
    waves[i] = new Wave();
    
  println("maxWaveRadius", maxWaveRadius);
}


void draw() {
 background(3, 26, 68);
 setView();
 
 for (int waveIdx = 0; waveIdx < waveCount; waveIdx++) {
   waves[waveIdx].update();
 }
 
 drawLines();
 
 if (savePicture) save(timestamp() + ".jpeg");
 savePicture = false;
 saveFrame("frames/####.png");
}


void drawLines() {
  noFill();
  strokeCap(ROUND);
  blendMode(BLEND);
  strokeWeight(lineWeight);
  
  for (int lineIdx = 0; lineIdx < lineCount; lineIdx++){
    float flatteningDistance = flatteningDistances[lineIdx];
    float y = map(lineIdx, 0, lineCount, 0, height) * flatteningDistance;
    PVector v, prev = new PVector(0, 0, 0);
    
    beginShape();
    for (int vertexIdx = 0; vertexIdx < lineDefinition; vertexIdx++){
      prev = vertexIdx == 0 ? new PVector(0, 0, 0) : prev;
      float x = map(vertexIdx, 0, lineDefinition - 1, 0, width);
      
      v = new PVector(x, y + lineHeightImperfections[lineIdx][vertexIdx], 0);
      addWaveComponent(v);
    
      color strokeColor = computeStrokeColor(v.z);
      stroke(strokeColor);
      
      float strokeWeight = computeStrokeWeight(prev, vertexIdx == 0 ? prev : v);
      strokeWeight(strokeWeight);
      
      curveVertex(v.x, v.y, v.z);
      if (vertexIdx == 0 || vertexIdx == lineDefinition - 1){
        curveVertex(v.x, v.y, v.z);
      }
      
      strokeWeight(1);
      prev = v;
    }
    endShape();
  }
}


color c1 = color(19, 123, 202);
//color c2 = color(28, 166, 233);
color c2 = color(43, 237, 247);
color c3 = color(39, 238, 110);
color c4 = color(253, 89, 233);

// Compute stroke color frome wave height
color computeStrokeColor(float vHeight) {
  float normalizedHeight = norm(vHeight, 0, maxWaveHeight / 8);
  float[] boundaries = new float[]{0.5, 0.8};
  return normalizedHeight < boundaries[0] ? 
    lerpColor(c1, c2, norm(normalizedHeight, 0, boundaries[0])) :
    normalizedHeight < boundaries[1] ? 
      lerpColor(c2, c3, norm(normalizedHeight, boundaries[0], boundaries[1])) :
      lerpColor(c3, c4, norm(normalizedHeight, boundaries[1], 1));
}


float computeStrokeWeight(PVector prev, PVector next) {
  if (prev.z <= next.z) return 1;
  PVector a = new PVector(prev.x, prev.z);
  PVector b = new PVector(next.x, next.z);
  float degs = angleBetweenPoints(a, b);
  float weight = map(degs, 90, 180, lineWeight, maxLineWeight);
  return weight;
}

void keyPressed(){
 if (key=='p' || key=='P') savePicture = true; 
}


// TODO Should be more natural
float[][] computeLineHeightImperfections() {
  float range = lineSpacing * lineHeightImperfectionRange;
  float[][] imperfections = new float[lineCount][lineDefinition];
  for (int i = 0; i < lineCount; i++) {
    for (int j = 0; j < lineDefinition * 4; j++) {
      float previous = j == 0 ? 0 : imperfections[i][j % lineDefinition];
      float next = floor(random(4)) == 0 ? random(-range, range) : 0;
      imperfections[i][j % lineDefinition] = previous + next;
      smoothNearby(imperfections[i], j, next);
    }
  }
  return imperfections;
}


void smoothNearby(float[] arr, int j, float b) {
  float a; int idx;
  if (j != 0){
    idx = (j - 1) % lineDefinition;
    a = arr[idx];
    arr[idx] = min(a, b) + abs(a - b) / 2;
  }
  if (j != lineDefinition){
    idx = (j + 1) % lineDefinition;
    a = arr[idx];
    arr[idx] = min(a, b) + abs(a - b) / 3;
  }
}


float[] computeFlatteningDistances() {
  float[] result = new float[lineCount];
  for (int i = 1; i <= lineCount; i++) {
    result[i - 1] = 1 + (lineCount - i) * 0.004;
    flatteningDistancesSum += result[lineCount - i]; 
  }
  return result;
}


void addWaveComponent(PVector linePoint){
  pushMatrix();
  scale(1, 0.1);
  for (int waveIdx = 0; waveIdx < waveCount; waveIdx++) {
    Wave wave = waves[waveIdx];
    float distance = sqrt(pow(wave.location.x - linePoint.x, 2) / 5 + pow(wave.location.y - linePoint.y, 2));
    float waveRadius = maxWaveRadius * abs(wave.currentStrength) * wave.maxRadiusCoeff ; // for any point in current time
    if (distance < waveRadius){
      float effectiveStrength = 1 - norm(distance, 0, waveRadius);
      float piStrength = map(effectiveStrength, 0, 1, 0, PI);
      float smoothed = -1 * cos(piStrength) + 1;
      linePoint.z += smoothed * wave.currentStrength * wave.maxAmplitudeCoeff * maxWaveHeight;
    }
  }
  popMatrix();
  float cappedZ = max(min(linePoint.z, maxWaveHeight), -maxWaveHeight);
  float normedZ = norm(abs(cappedZ), 0, maxWaveHeight);
  linePoint.z = cappedZ * (-1 * sqrt(normedZ) + 1);
}


void setView() {
  scale(1.25);
  //translate(-190, -170, -260);
  translate(-210, -190, -260);
  rotateX(rotationX); 
  rotateY(rotationY); 
  rotateZ(rotationZ);
}

float randomPlusMinus(float base, float delta) {
  return random(base - delta, base + delta);
}


class Wave {
  PVector location, velocity;
  float currentStrength; // (-1, 1);
  float maxAmplitudeCoeff = 0.05; // of maxWaveHeight
  float maxRadiusCoeff = 0.15; // of maxWaveRadius
  float halfLife;
  int age;

  Wave() {
    this.reset(true);
  }

  void update() {
    currentStrength = sin(map(age, 0, halfLife * 4, 0, 2*PI));
    location = PVector.add(location, velocity);  
    age++;
    if (currentStrength < 0.01 && location.x > 1.5 * width){
      reset(false);
    }
    //ellipse(location.x, location.y, 30, 30); // for debugging
  }
  
  void reset(boolean isNew) {
    location = new PVector(random(-width*1.5, isNew ? width : 0), random(0, height), 0);
    velocity = new PVector(width * 0.1 * randomPlusMinus(waveVelocity, 0.01), 0, 0);
    maxAmplitudeCoeff = random(20) == 0 ? 0.2 : random(0.05, 0.07);
    maxRadiusCoeff = random(20) == 0 ? 0.3 : random(0.1, 0.14);
    halfLife = random(100, 120);
    //if (!isNew) age = 0
    age = floor(random(0, halfLife * 4));
  }
}


String timestamp() {
  return String.format("%1$ty%1$tm%1$td_%1$tH%1$tM%1$tS", Calendar.getInstance());
}

float angleBetweenPoints(PVector a, PVector b){
   return degrees(atan2((b.x - a.x), (b.y - a.y)));
}